#!/usr/bin/perl
# bin2dump
# V0.3
#
# Copyright(c) 2012 Sasaji
#
use File::Basename;
use Getopt::Long;

my $debug = 0;
my %optvalues = (
	"outfile" => "",
	"help"    => "",
	"noheader" => "",
	"startaddr" => "",
);

my $result = GetOptions("outfile=s" => \$optvalues{"outfile"},
                        "help"      => \$optvalues{"help"},
                        "noheader"  => \$optvalues{"noheader"},
                        "startaddr=s" => \$optvalues{"startaddr"},
                        );
if ($optvalues{"help"} || @ARGV <= 0) {
	usage();
	exit 0;
}
my $outbuf="";
my $outbuflen = 0;

foreach my $file (@ARGV) {
	read_file($file);
}

my $addr_start = 0;

if (!$optvalues{"noheader"}) {
	my $s = substr($outbuf,0,5);
	my $e = substr($outbuf,-5);
	$outbuf = substr($outbuf,5,$outbuflen-10);
	$outbuflen -= 10;
	$addr_start = unpack("n",substr($s,3,2));
}

if ($optvalues{"startaddr"}) {
	my $s = unpack("n",pack("H4", $optvalues{"startaddr"}));
	$addr_start = $s;
}

# output filename
my $outfile = "";
if ($optvalues{"outfile"} ne "") {
	$outfile = $optvalues{"outfile"};
#} else {
#	$outfile = $ARGV[0];
#	$outfile =~ s/\.(.+?)$//;
#	$outfile .= ".txt";
}

my $addr = $addr_start;
out_dump();

print STDERR "Complete.\n";
exit 0;

sub read_file() {
	my($infile)=@_;
	my $fdin;

	my $buf;
	my $buflen = 0;

	if (!open($fdin,$infile)) {
		print STDERR "Warning: $infile doesn't found.\n";
		return 1;
	} else {
		print STDERR "<<< $infile >>>\n";
	}
	binmode($fdin);
	while (($buflen = read($fdin,$buf,256)) > 0) {
		$outbuf .= $buf;
		$outbuflen += $buflen;
	}
	close($dfin);
}

sub out_dump() {
	my $fdout;

	my $x;
	my $y;
	my $buf;
	my $pos = 0;
	my $data = 0;
	my $sum1 = 0;
	my @sumcol = ();

	if ($outfile ne "") {
		if (!open($fdout,"> ".$outfile)) {
			print STDERR "Error: $outfile cannot open.\n";
			return 1;
		} else {
			print STDERR "Output: $outfile\n";
		}
	} else {
		$fdout = STDOUT;
	}

	for($pos = 0; $pos < (int(($outbuflen-1) / 256)+1)*256; ) {
		@sumcol = ();
		print {$fdout} "Add: +0 +1 +2 +3 +4 +5 +6 +7 +8 +9 +A +B +C +D +E +F Sum\n";
		for($y = 0; $y < 16; $y++) {
			printf {$fdout} "%04X",$addr;
			$sum1 = 0;
			for($x = 0; $x < 16; $x++) {
				if ($pos < $outbuflen) {
					$data = unpack("C",substr($outbuf,$pos,1));
				} else {
					$data = 0;
				}

				printf {$fdout} " %02X",$data;

				$sum1 += $data;
				$sumcol[$x] += $data;
				$pos++;
				$addr++;
			}
			printf {$fdout} " :%02X",($sum1 % 256);
			print {$fdout} "\n";
		}
		print {$fdout} "---------------------------------------------------------\n";
		print {$fdout} "Sum:";
		$sum1 = 0;
		for($x = 0; $x < 16; $x++) {
			printf {$fdout} " %02X",($sumcol[$x] % 256);
			$sum1 += $sumcol[$x];
		}
		printf {$fdout} " :%02X",($sum1 % 256);
		print {$fdout} "\n";
		print {$fdout} "\n";
	}
	if ($outfile ne "") {
		close($fdout);
	}
}

sub usage
{
	my $prog = basename($0);
	print STDERR "Usage: $prog [<options>] <binary file...>\n";
	print STDERR " options: -o <outputfile> : output file name\n";
	print STDERR "          -n : no include header in specified binary file\n";
	print STDERR "          -s <address(4chars hex)> : specify start address\n";
}
