#!/usr/bin/perl
# dump2bin
# V0.3
#
# Copyright(c) 2012 Sasaji
#
use File::Basename;
use Getopt::Long;

my $debug = 0;
my %optvalues = (
	"outfile" => "",
	"help"    => "",
	"noheader" => "",
	"startaddr" => "",
	"execaddr" => "",
);

my $result = GetOptions("outfile=s" => \$optvalues{"outfile"},
                        "help"      => \$optvalues{"help"},
                        "noheader"  => \$optvalues{"noheader"},
                        "startaddr=s" => \$optvalues{"startaddr"},
                        "execaddr=s"  => \$optvalues{"execaddr"},
                        );
if ($optvalues{"help"} || @ARGV <= 0) {
	usage();
	exit 0;
}

my $addr_total_start = -1;
my $addr_total_end = -1;
my $addr_total_exec = -1;
my $addr_total_size = -1;
my $outbuf="";

# output filename
my $outfile = "";
if ($optvalues{"outfile"} ne "") {
	$outfile = $optvalues{"outfile"};
#} else {
#	$outfile = $ARGV[0];
#	$outfile =~ s/\.(.+?)$//;
#	$outfile .= ".bin";
}

foreach my $file (@ARGV) {
	parse($file);
}

if ($optvalues{"startaddr"}) {
	my $s = unpack("n",pack("H4", $optvalues{"startaddr"}));
	my $n = $s - $addr_total_start;
	$addr_total_start += $n;
	$addr_total_end   += $n;
}
if ($optvalues{"execaddr"}) {
	my $e = unpack("n",pack("H4", $optvalues{"execaddr"}));
	$addr_total_exec = $e;
} else {
	$addr_total_exec = $addr_total_start;
}
$addr_total_size = $addr_total_end-$addr_total_start;

print STDERR "\n";
printf STDERR "Total: \$%04x - \$%04x (size: \$%04x) exec: \$%04x\n",$addr_total_start,$addr_total_end,$addr_total_size,$addr_total_exec;

# add header for outfile
if (!$optvalues{"noheader"}) {
	$outbuf = "\x00".pack("nn",$addr_total_size,$addr_total_start).$outbuf;
	$outbuf .= "\xff\x00\x00".pack("n",$addr_total_exec);
}

if ($outfile ne "") {
	if (!open($fdout,"> ".$outfile)) {
		print STDERR "Error: $outfile cannot open.\n";
		exit 1;
	} else {
		print STDERR "Output: $outfile\n";
	}
} else {
	$fdout = STDOUT;
}
binmode($fdout);
print {$fdout} $outbuf;
if ($outfile ne "") {
	close($fdout);
}
print STDERR "Complete.\n";
exit 0;



sub parse() {
	my($infile)=@_;
	my $fdin;
	my $fdout;
	my $line;
	my $linenum = 0;
	if (!open($fdin,$infile)) {
		print STDERR "Warning: $infile doesn't found.\n";
		return 1;
	} else {
		print STDERR "<<< $infile >>>\n";
	}

	my $datas = [];
	my $addr = [];
	my $addr_start = -1;
	my $addr_end = -1;
	my $sum_all;
	my $sum1col;
	my $sum1raw;
	my $x;
	my $y = 0;
	while ( $line = readline($fdin) ) {
		$linenum++;
		if ($line =~ /^([0-9a-fA-F]{4})\s+/) {
			$addr->[$y] = hex($1);
			$line = $';	# post match

			$addr_start = $addr->[$y] if ($addr_start < 0);
			$addr_end = $addr->[$y];
			$addr_total_start = $addr->[$y] if ($addr_total_start < 0);

			printf("%04x ",$addr->[$y]) if ($debug);
			if ($y > 0 && ($addr->[$y] - $addr->[$y-1]) != 16) {
				printf STDERR ("Warning: address isn't continuous: \$%04x(line %d).\n",$addr->[$y], $linenum);
			}

			for($x=0; $x<17; $x++) {
				$datas->[$y] = [] if (! defined($datas->[$y]));

				if ($line =~ /^([0-9a-fA-F]{2})\s+/) {
					$datas->[$y][$x] = hex($1);
					$line = $';	# post match

					printf STDERR ("%02x    ",$datas->[$y][$x]) if ($debug);
					$outbuf .= pack("C", $datas->[$y][$x]);

				} elsif ($line =~ /^:([0-9a-fA-F]{2})/) {
					$datas->[$y][$x] = hex($1);

					printf STDERR (":%02x",$datas->[$y][$x]) if ($debug);

				} else {
					printf STDERR ("Error: unknown char exist at line %d. (%s)\n"
					,$linenum
					,$line);

					last;
				}
			}
			$sum1raw = 0;
			for($xx=0; $xx<16; $xx++) {
				$sum1raw += $datas->[$y][$xx];
				$addr_end++;
			}
			printf STDERR (":%02x\n",($sum1raw & 0xff)) if ($debug);
			if (($sum1raw & 0xff) != $datas->[$y][16]) {
				printf STDERR ("Error: checksum doesn't match at \$%04x(line %d). sum:%02x data:%02x\n",$addr->[$y],$linenum,($sum1raw & 0xff),$datas->[$y][16]);
			}
			$y++;

		} elsif (lc($line) =~ /^sum:\s+/) {
			$line = $';	# post match

			print STDERR ("SUM: ") if ($debug);

			for($x=0; $x<17; $x++) {
				$datas->[$y] = [] if (! defined($datas->[$y]));

				if ($line =~ /^([0-9a-fA-F]{2})\s+/) {
					$datas->[$y][$x] = hex($1);
					$line = $';	# post match

					$sum1col = 0;
					for($yy=0; $yy<$y; $yy++) {
						$sum1col += $datas->[$yy][$x];
					}

					printf STDERR ("%02x:%02x ",$datas->[$y][$x],($sum1col & 0xff)) if ($debug);
					if (($sum1col & 0xff) != $datas->[$y][$x]) {
						printf STDERR ("Error: checksum doesn't match in \$%04x-\$%04x(line %d), column %x. sum:%02x data:%02x\n"
						,$addr->[0],$addr->[$y-1],$linenum,$x
						,($sum1col & 0xff),$datas->[$y][$x]);
					}

				} elsif ($line =~ /^:([0-9a-fA-F]{2})/) {
					$datas->[$y][$x] = hex($1);

					printf STDERR (":%02x",$datas->[$y][$x]) if ($debug);

				} else {
					printf STDERR ("Error: unknown char exists at line %d. (%s)\n"
					,$linenum
					,$line);

					last;
				}
			}
			$sum_all = 0;
			for($xx=0; $xx<16; $xx++) {
				$sum_all += $datas->[$y][$xx];
			}
			printf STDERR (":%02x\n",($sum_all & 0xff)) if ($debug);
			if (($sum_all & 0xff) != $datas->[$y][16]) {
				printf STDERR ("Error: checksum total doesn't match in \$%04x-\$%04x(line %d). sum:%02x data:%02x\n"
				,$addr->[0],$addr->[$y-1],$linenum,
				,($sum_all & 0xff),$datas->[$y][16]);
			}

			$y=0;
			$datas = [];
			$addr = [];


		}
	}

	close($fdin);

	$addr_total_end = $addr_end;

	printf STDERR ("Addr: \$%04x - \$%04x (size: \$%04x)\n",$addr_start,$addr_end,$addr_end-$addr_start);

}

sub usage
{
	my $prog = basename($0);
	print STDERR "Usage: $prog [<options>] <dump text file...>\n";
	print STDERR " options: -o <outputfile> : output file name\n";
	print STDERR "          -n : add no header\n";
	print STDERR "          -s <address(4chars hex)> : specify start address\n";
	print STDERR "          -e <address(4chars hex)> : specify execute start address\n";
}
